import dataclasses

import redis
from redis.asyncio import Redis

DEFAULT_CONNECTION = "redis://localhost:6379/0"


def create_all_types():
    from rapyer.types.init import ALL_TYPES

    return ALL_TYPES


@dataclasses.dataclass
class RedisConfig:
    redis: Redis = dataclasses.field(
        default_factory=lambda: redis.asyncio.from_url(
            DEFAULT_CONNECTION, decode_responses=True
        )
    )
    redis_type: dict[type, type] = dataclasses.field(default_factory=create_all_types)
    ttl: int | None = None



from typing import TypeAlias, TYPE_CHECKING

from pydantic_core import core_schema
from pydantic_core.core_schema import ValidationInfo, SerializationInfo

from rapyer.types.base import RedisType, REDIS_DUMP_FLAG_NAME


class RedisBytes(bytes, RedisType):
    original_type = bytes

    def clone(self):
        return bytes(self)

    def __iadd__(self, other):
        new_value = self + other
        if self.pipeline:
            self.pipeline.json().set(self.key, self.json_path, new_value)
        return self.__class__(new_value)

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler):
        return core_schema.no_info_after_validator_function(
            cls,
            core_schema.with_info_before_validator_function(
                cls._validate_pickle, handler(cls.original_type)
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls._serialize_pickle,
                return_schema=core_schema.str_schema(),
                info_arg=True,
            ),
        )

    @classmethod
    def _validate_pickle(cls, value, info: ValidationInfo):
        ctx = info.context or {}
        is_redis_data = ctx.get(REDIS_DUMP_FLAG_NAME)

        if isinstance(value, str) and is_redis_data:
            return cls.deserialize_unknown(value)
        return value

    @classmethod
    def _serialize_pickle(cls, value, info: SerializationInfo):
        ctx = info.context or {}
        is_redis_data = ctx.get(REDIS_DUMP_FLAG_NAME)

        value = bytes(value)
        if is_redis_data:
            return cls.serialize_unknown(value)
        return value


if TYPE_CHECKING:
    RedisBytes: TypeAlias = RedisBytes | bytes



import dataclasses
from typing import Annotated


@dataclasses.dataclass(frozen=True)
class KeyAnnotation:
    pass


def Key(typ: type = None):
    if typ is None:
        return KeyAnnotation()

    return Annotated[typ, KeyAnnotation()]



from typing import TypeAlias, TYPE_CHECKING

from rapyer.types.base import RedisType


class RedisStr(str, RedisType):
    original_type = str

    def clone(self):
        return str(self)

    def __iadd__(self, other):
        new_value = self + other
        if self.pipeline:
            self.pipeline.json().set(self.key, self.json_path, new_value)
        return self.__class__(new_value)


if TYPE_CHECKING:
    RedisStr: TypeAlias = RedisStr | str



import redis.asyncio as redis_async
from redis.asyncio.client import Redis

from rapyer.base import REDIS_MODELS


async def init_rapyer(redis: str | Redis = None, ttl: int = None):
    if isinstance(redis, str):
        redis = redis_async.from_url(redis, decode_responses=True, max_connections=20)

    for model in REDIS_MODELS:
        if redis is not None:
            model.Meta.redis = redis
        if ttl is not None:
            model.Meta.ttl = ttl


async def teardown_rapyer():
    for model in REDIS_MODELS:
        if model.Meta.ttl is not None:
            await model.Meta.redis.aclose()



import json
from typing import TypeVar, TYPE_CHECKING

from pydantic_core import core_schema
from pydantic_core.core_schema import ValidationInfo, SerializationInfo
from typing_extensions import TypeAlias

from rapyer.types.base import GenericRedisType, RedisType, REDIS_DUMP_FLAG_NAME

T = TypeVar("T")


class RedisList(list, GenericRedisType[T]):
    original_type = list

    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        GenericRedisType.__init__(self, *args, **kwargs)

    def create_new_values(self, keys, values):
        new_values = self._adapter.validate_python(values)
        for key, value in zip(keys, new_values):
            self.init_redis_field(f"[{key}]", value)
        return new_values

    def create_new_value(self, key, value):
        new_val = self.create_new_values([key], [value])[0]
        return new_val

    def __setitem__(self, key, value):
        if self.pipeline:
            self.pipeline.json().set(self.key, self.json_field_path(key), value)
        new_val = self.create_new_value(key, value)
        return super().__setitem__(key, new_val)

    def __iadd__(self, other):
        self.extend(other)
        if self.pipeline and other:
            self.pipeline.json().arrappend(self.key, self.json_path, *other)
        return self

    def append(self, __object):
        if self.pipeline:
            self.pipeline.json().arrappend(self.key, self.json_path, __object)
        key = len(self)
        new_val = self.create_new_value(key, __object)
        return super().append(new_val)

    def extend(self, new_lst):
        if self.pipeline and new_lst:
            self.pipeline.json().arrappend(self.key, self.json_path, *new_lst)
        new_keys = range(len(self), len(self) + len(new_lst))
        new_vals = self.create_new_values(list(new_keys), new_lst)
        return super().extend(new_vals)

    def insert(self, index, __object):
        if self.pipeline:
            self.pipeline.json().arrinsert(self.key, self.json_path, index, __object)
        new_val = self.create_new_value(index, __object)
        return super().insert(index, new_val)

    def clear(self):
        if self.pipeline:
            self.pipeline.json().set(self.key, self.json_path, [])
        return super().clear()

    async def aappend(self, __object):
        self.append(__object)

        # Serialize the object for Redis storage using a type adapter
        if not self.pipeline:
            serialized_object = self._adapter.dump_python(
                [__object], mode="json", context={REDIS_DUMP_FLAG_NAME: True}
            )
            await self.redis.json().arrappend(
                self.key, self.json_path, *serialized_object
            )

    async def aextend(self, __iterable):
        items = list(__iterable)
        self.extend(items)

        # Convert iterable to list and serialize items using type adapter
        if items and not self.pipeline:
            # normalized_items = self._adapter.validate_python(items)
            serialized_items = self._adapter.dump_python(
                items, mode="json", context={REDIS_DUMP_FLAG_NAME: True}
            )
            await self.redis.json().arrappend(
                self.key,
                self.json_path,
                *serialized_items,
            )

    async def apop(self, index=-1):
        if self:
            self.pop(index)
        arrpop = await self.redis.json().arrpop(self.key, self.json_path, index)

        # Handle empty list case
        if arrpop is None or (isinstance(arrpop, list) and len(arrpop) == 0):
            return None

        # Handle case where arrpop returns [None] for an empty list
        if arrpop[0] is None:
            return None
        arrpop = [json.loads(val) for val in arrpop]
        return self._adapter.validate_python(
            arrpop, context={REDIS_DUMP_FLAG_NAME: True}
        )[0]

    async def ainsert(self, index, __object):
        self.insert(index, __object)

        if not self.pipeline:
            # Serialize the object for Redis storage using a type adapter
            serialized_object = self._adapter.dump_python(
                [__object], mode="json", context={REDIS_DUMP_FLAG_NAME: True}
            )
            await self.redis.json().arrinsert(
                self.key, self.json_path, index, *serialized_object
            )

    async def aclear(self):
        # Clear local list
        self.clear()

        # Clear Redis list
        if not self.pipeline:
            await self.client.json().set(self.key, self.json_path, [])

    def clone(self):
        return [v.clone() if isinstance(v, RedisType) else v for v in self]

    def iterate_items(self):
        keys = [f"[{i}]" for i in range(len(self))]
        return zip(keys, self)

    @classmethod
    def full_serializer(cls, value, info: SerializationInfo):
        ctx = info.context or {}
        is_redis_data = ctx.get(REDIS_DUMP_FLAG_NAME)
        return [
            cls.serialize_unknown(item) if is_redis_data else item for item in value
        ]

    @classmethod
    def full_deserializer(cls, value, info: ValidationInfo):
        ctx = info.context or {}
        is_redis_data = ctx.get(REDIS_DUMP_FLAG_NAME)

        if isinstance(value, list):
            return [
                cls.deserialize_unknown(item) if is_redis_data else item
                for item in value
            ]
        return value

    @classmethod
    def schema_for_unknown(cls):
        core_schema.list_schema(core_schema.str_schema())


if TYPE_CHECKING:
    RedisList: TypeAlias = RedisList[T] | list[T]



import asyncio
import contextlib
import uuid
from datetime import timedelta

from redis.asyncio import Redis


@contextlib.asynccontextmanager
async def acquire_lock(
    redis: Redis, key: str, lock_timeout: timedelta | int = 10, sleep_time: int = 0.1
):
    lock_key = f"{key}:lock"
    lock_token = str(uuid.uuid4())
    while not await redis.set(lock_key, lock_token, nx=True, ex=lock_timeout):
        await asyncio.sleep(sleep_time)
    try:
        yield
    finally:
        await redis.delete(lock_key)


def update_keys_in_pipeline(pipeline, redis_key: str, **kwargs):
    for json_path, value in kwargs.items():
        pipeline.json().set(redis_key, json_path, value)



"""Redis Pydantic - Pydantic models with Redis as the backend."""

from rapyer.base import AtomicRedisModel, get, find_redis_models
from rapyer.init import init_rapyer, teardown_rapyer

__all__ = [
    "AtomicRedisModel",
    "init_rapyer",
    "teardown_rapyer",
    "get",
    "find_redis_models",
]



from typing import Any, get_origin

from pydantic import BaseModel, PrivateAttr, TypeAdapter

from rapyer.types.base import RedisType
from rapyer.utils.annotation import TypeConverter, DYNAMIC_CLASS_DOC
from rapyer.utils.pythonic import safe_issubclass


class RedisConverter(TypeConverter):
    def __init__(self, supported_types: dict[type, type], field_name: str):
        self.supported_types = supported_types
        self.field_name = field_name

    def is_redis_type(self, type_to_check: type) -> bool:
        origin = get_origin(type_to_check) or type_to_check
        if safe_issubclass(origin, RedisType):
            return True
        from rapyer.base import AtomicRedisModel

        return safe_issubclass(origin, AtomicRedisModel)

    def is_type_support(self, type_to_check: type) -> bool:
        if safe_issubclass(type_to_check, BaseModel):
            return True
        if safe_issubclass(type_to_check, RedisType):
            return True
        return type_to_check in self.supported_types

    def convert_flat_type(self, type_to_convert: type) -> type:
        if type_to_convert is Any:
            return Any

        from rapyer.base import AtomicRedisModel

        if safe_issubclass(type_to_convert, AtomicRedisModel):
            return type(
                type_to_convert.__name__,
                (type_to_convert,),
                dict(
                    _field_name=PrivateAttr(default=self.field_name),
                    __doc__=DYNAMIC_CLASS_DOC,
                ),
            )
        if safe_issubclass(type_to_convert, BaseModel):
            origin: type[BaseModel]
            return type(
                f"Redis{type_to_convert.__name__}",
                (AtomicRedisModel, type_to_convert),
                dict(
                    _field_name=PrivateAttr(default=self.field_name),
                    __doc__=DYNAMIC_CLASS_DOC,
                ),
            )
        if safe_issubclass(type_to_convert, RedisType):
            redis_type = type_to_convert
            original_type = type_to_convert.original_type
        else:
            redis_type = self.supported_types[type_to_convert]
            original_type = type_to_convert

        new_type = type(
            redis_type.__name__,
            (redis_type,),
            dict(
                field_name=self.field_name,
                original_type=original_type,
                __doc__=DYNAMIC_CLASS_DOC,
            ),
        )

        new_type._adapter = TypeAdapter(new_type)
        return new_type

    def covert_generic_type(
        self, type_to_covert: type, generic_values: tuple[type]
    ) -> type:
        from rapyer.base import AtomicRedisModel

        if safe_issubclass(type_to_covert, AtomicRedisModel):
            return type(
                type_to_covert.__name__,
                (type_to_covert,),
                dict(
                    _field_name=PrivateAttr(default=self.field_name),
                    __doc__=DYNAMIC_CLASS_DOC,
                ),
            )
        if safe_issubclass(type_to_covert, BaseModel):
            type_to_covert: type[BaseModel]
            return type(
                f"Redis{type_to_covert.__name__}",
                (AtomicRedisModel, type_to_covert),
                dict(
                    _field_name=PrivateAttr(default=self.field_name),
                    __doc__=DYNAMIC_CLASS_DOC,
                ),
            )

        if safe_issubclass(type_to_covert, RedisType):
            redis_type = type_to_covert
            original_type = type_to_covert.original_type
        else:
            redis_type = self.supported_types[type_to_covert]
            original_type = type_to_covert
            original_type = original_type[generic_values]

        new_type = type(
            redis_type.__name__,
            (redis_type,),
            dict(
                field_name=self.field_name,
                original_type=original_type,
                __doc__=DYNAMIC_CLASS_DOC,
            ),
        )
        adapter_type = new_type[generic_values]

        if issubclass(redis_type, RedisType):
            new_type._adapter = TypeAdapter(adapter_type)
        return new_type[generic_values]



from datetime import datetime
from typing import TYPE_CHECKING

from rapyer.types.base import RedisType


class RedisDatetime(datetime, RedisType):
    original_type = datetime

    def __new__(cls, value, *args, **kwargs):
        if isinstance(value, datetime):
            # Support init from a datetime, preserving microseconds and timezone
            return datetime.__new__(
                cls,
                value.year,
                value.month,
                value.day,
                value.hour,
                value.minute,
                value.second,
                value.microsecond,
                value.tzinfo,
            )
        else:
            return datetime.__new__(cls, value, *args, **kwargs)

    def clone(self):
        return datetime.fromtimestamp(self.timestamp())


if TYPE_CHECKING:
    RedisDatetime = RedisDatetime | datetime



from datetime import datetime

from rapyer.types.byte import RedisBytes
from rapyer.types.datetime import RedisDatetime
from rapyer.types.dct import RedisDict
from rapyer.types.integer import RedisInt
from rapyer.types.lst import RedisList
from rapyer.types.string import RedisStr

ALL_TYPES = {
    list: RedisList,
    dict: RedisDict,
    bytes: RedisBytes,
    int: RedisInt,
    str: RedisStr,
    datetime: RedisDatetime,
}
